@@ -936,6 +936,9 @@ function test_vector_nonlinear_oracle()
         @test 0.9 < cache.eval_jacobian_timer < 4
         @test 0.9 < cache.eval_hessian_lagrangian_timer < 2
     end
+    MOI.set(model, MOI.RawOptimizerAttribute("max_iter"), 0)
+    MOI.optimize!(model)
+    @test MOI.get(model, MOI.PrimalStatus()) == MOI.INFEASIBLE_POINT
     return
 end
 
@@ -1240,6 +1243,15 @@ function test__VectorNonlinearOracle()
     return
 end
 
+function test_default_objective_function()
+    model = Ipopt.Optimizer()
+    F = MOI.get(model, MOI.ObjectiveFunctionType())
+    @test F == MOI.ScalarAffineFunction{Float64}
+    f = MOI.get(model, MOI.ObjectiveFunction{F}())
+    @test f ≈ zero(F)
+    return
+end
+
 end  # module TestMOIWrapper
 
 TestMOIWrapper.runtests()
